@@ -31,6 +31,7 @@
 #include <mntent.h>
 
 #include <sys/ioctl.h>
+#include <sys/file.h>
 #include <linux/fs.h>
 
 #include <string.h>
@@ -53,14 +54,12 @@
 #include "udiskslinuxswapspace.h"
 #include "udiskslinuxloop.h"
 #include "udiskslinuxprovider.h"
-#include "udisksfstabmonitor.h"
-#include "udisksfstabentry.h"
 #include "udiskscrypttabmonitor.h"
 #include "udiskscrypttabentry.h"
 #include "udiskslinuxdevice.h"
 #include "udisksmodulemanager.h"
-
-#include <modules/udisksmoduleifacetypes.h>
+#include "udisksmodule.h"
+#include "udisksmoduleobject.h"
 
 /**
  * SECTION:udiskslinuxblockobject
@@ -86,6 +85,9 @@ struct _UDisksLinuxBlockObject
   UDisksMountMonitor *mount_monitor;
 
   UDisksLinuxDevice *device;
+  GMutex device_mutex;
+
+  GMutex cleanup_mutex;
 
   /* interface */
   UDisksBlock *iface_block_device;
@@ -103,14 +105,6 @@ struct _UDisksLinuxBlockObjectClass
   UDisksObjectSkeletonClass parent_class;
 };
 
-typedef struct
-{
-  UDisksObject *interface;
-  UDisksObjectHasInterfaceFunc has_func;
-  UDisksObjectConnectInterfaceFunc connect_func;
-  UDisksObjectUpdateInterfaceFunc update_func;
-} ModuleInterfaceEntry;
-
 enum
 {
   PROP_0,
@@ -133,10 +127,13 @@ udisks_linux_block_object_finalize (GObject *_object)
   UDisksLinuxBlockObject *object = UDISKS_LINUX_BLOCK_OBJECT (_object);
 
   /* note: we don't hold a ref to block->daemon or block->mount_monitor */
-  g_signal_handlers_disconnect_by_func (object->mount_monitor, on_mount_monitor_mount_added, object);
-  g_signal_handlers_disconnect_by_func (object->mount_monitor, on_mount_monitor_mount_removed, object);
+  g_warn_if_fail (g_signal_handlers_disconnect_by_func (object->mount_monitor, on_mount_monitor_mount_added, object) == 1);
+  g_warn_if_fail (g_signal_handlers_disconnect_by_func (object->mount_monitor, on_mount_monitor_mount_removed, object) == 1);
 
   g_object_unref (object->device);
+  g_mutex_clear (&object->device_mutex);
+
+  g_mutex_clear (&object->cleanup_mutex);
 
   if (object->iface_block_device != NULL)
     g_object_unref (object->iface_block_device);
@@ -222,6 +219,12 @@ udisks_linux_block_object_constructed (GObject *_object)
   UDisksLinuxBlockObject *object = UDISKS_LINUX_BLOCK_OBJECT (_object);
   GString *str;
   UDisksBlock *block = NULL;
+  UDisksPartition *partition = NULL;
+
+  g_mutex_init (&object->device_mutex);
+  g_mutex_init (&object->cleanup_mutex);
+
+  object->module_ifaces = g_hash_table_new_full (g_direct_hash, g_direct_equal, NULL, g_object_unref);
 
   object->mount_monitor = udisks_daemon_get_mount_monitor (object->daemon);
   g_signal_connect (object->mount_monitor,
@@ -246,6 +249,13 @@ udisks_linux_block_object_constructed (GObject *_object)
   if (block && g_strcmp0 (udisks_block_get_crypto_backing_device (block), "/") != 0)
     udisks_linux_block_object_uevent (object, "change", NULL);
 
+  /* run update for partitions again -- it sets the "Partitions" property
+   * on the PartitionTable interface and we need the object path for that
+   */
+  partition = udisks_object_peek_partition (UDISKS_OBJECT (object));
+  if (partition)
+    udisks_linux_block_object_uevent (object, "change", NULL);
+
   if (G_OBJECT_CLASS (udisks_linux_block_object_parent_class)->constructed != NULL)
     G_OBJECT_CLASS (udisks_linux_block_object_parent_class)->constructed (_object);
 }
@@ -344,8 +354,15 @@ udisks_linux_block_object_get_daemon (UDisksLinuxBlockObject *object)
 UDisksLinuxDevice *
 udisks_linux_block_object_get_device (UDisksLinuxBlockObject *object)
 {
+  UDisksLinuxDevice *device;
+
   g_return_val_if_fail (UDISKS_IS_LINUX_BLOCK_OBJECT (object), NULL);
-  return g_object_ref (object->device);
+
+  g_mutex_lock (&object->device_mutex);
+  device = g_object_ref (object->device);
+  g_mutex_unlock (&object->device_mutex);
+
+  return device;
 }
 
 /**
@@ -374,6 +391,31 @@ udisks_linux_block_object_get_device_file (UDisksLinuxBlockObject *object)
   return device_file;
 }
 
+/**
+ * udisks_linux_block_object_get_device_number:
+ * @object: A #UDisksLinuxBlockObject.
+ *
+ * Gets the device number for this object.
+ *
+ * Returns: A dev_t device number or zero in case of an error.
+ */
+dev_t
+udisks_linux_block_object_get_device_number (UDisksLinuxBlockObject *object)
+{
+  UDisksLinuxDevice *device;
+  dev_t device_number;
+
+  g_return_val_if_fail (UDISKS_IS_LINUX_BLOCK_OBJECT (object), 0);
+
+  device = udisks_linux_block_object_get_device (object);
+  device_number = g_udev_device_get_device_number (device->udev_device);
+
+  /* Free resources. */
+  g_object_unref (device);
+
+  return device_number;
+}
+
 /* ---------------------------------------------------------------------------------------------------- */
 
 static void
@@ -404,6 +446,7 @@ update_iface (UDisksObject                     *object,
       if (has)
         {
           *interface_pointer = g_object_new (skeleton_type, NULL);
+          g_warn_if_fail (*interface_pointer != NULL);
           if (connect_func != NULL)
             connect_func (object);
           add = TRUE;
@@ -507,8 +550,8 @@ partition_table_check (UDisksObject *object)
        * mark us as a partition table ... except if we are partitioned by the
        * kernel
        *
-       * (see filesystem_check() for the similar case where we don't pretend
-       * to be a filesystem)
+       * (see udisks_linux_block_object_contains_filesystem() for the similar case
+       * where we don't pretend to be a filesystem)
        */
       if (g_strcmp0 (g_udev_device_get_property (block_object->device->udev_device, "ID_FS_USAGE"), "filesystem") == 0)
         {
@@ -619,10 +662,10 @@ drive_does_not_detect_media_change (UDisksLinuxBlockObject *object)
   return ret;
 }
 
+
 static gboolean
-filesystem_check (UDisksObject *object)
+contains_filesystem_locked (UDisksLinuxBlockObject *block_object, UDisksLinuxDevice *device)
 {
-  UDisksLinuxBlockObject *block_object = UDISKS_LINUX_BLOCK_OBJECT (object);
   gboolean ret = FALSE;
   gboolean detected_as_filesystem = FALSE;
   UDisksMountType mount_type;
@@ -637,8 +680,8 @@ filesystem_check (UDisksObject *object)
        * (see partition_table_check() above for the similar case where we don't pretend
        * to be a partition table)
        */
-      if (g_strcmp0 (g_udev_device_get_devtype (block_object->device->udev_device), "disk") == 0 &&
-          disk_is_partitioned_by_kernel (block_object->device->udev_device))
+      if (g_strcmp0 (g_udev_device_get_devtype (device->udev_device), "disk") == 0 &&
+          disk_is_partitioned_by_kernel (device->udev_device))
         {
           detected_as_filesystem = FALSE;
         }
@@ -647,7 +690,7 @@ filesystem_check (UDisksObject *object)
   if (drive_does_not_detect_media_change (block_object) ||
       detected_as_filesystem ||
       (udisks_mount_monitor_is_dev_in_use (block_object->mount_monitor,
-                                           g_udev_device_get_device_number (block_object->device->udev_device),
+                                           g_udev_device_get_device_number (device->udev_device),
                                            &mount_type) &&
        mount_type == UDISKS_MOUNT_TYPE_FILESYSTEM))
     ret = TRUE;
@@ -655,6 +698,43 @@ filesystem_check (UDisksObject *object)
   return ret;
 }
 
+static gboolean
+contains_filesystem (UDisksObject *object)
+{
+  UDisksLinuxBlockObject *block_object = UDISKS_LINUX_BLOCK_OBJECT (object);
+
+  return contains_filesystem_locked (block_object, block_object->device);
+}
+
+/**
+ * udisks_linux_block_object_contains_filesystem:
+ * @object: A #UDisksLinuxBlockObject.
+ *
+ * Check whether filesystem has been detected on the block device.
+ *
+ * Returns: %TRUE if the block device contains filesystem, %FALSE otherwise.
+ */
+gboolean
+udisks_linux_block_object_contains_filesystem (UDisksObject *object)
+{
+  UDisksLinuxBlockObject *block_object;
+  UDisksLinuxDevice *device;
+  gboolean ret;
+
+  g_return_val_if_fail (UDISKS_IS_LINUX_BLOCK_OBJECT (object), FALSE);
+  block_object = UDISKS_LINUX_BLOCK_OBJECT (object);
+
+  g_mutex_lock (&block_object->device_mutex);
+  device = g_object_ref (block_object->device);
+  g_mutex_unlock (&block_object->device_mutex);
+
+  ret = contains_filesystem_locked (block_object, device);
+
+  g_object_unref (device);
+
+  return ret;
+}
+
 
 static void
 filesystem_connect (UDisksObject *object)
@@ -715,6 +795,7 @@ encrypted_check (UDisksObject *object)
 
   return udisks_linux_block_is_luks (block) ||
          udisks_linux_block_is_tcrypt (block) ||
+         udisks_linux_block_is_bitlk (block) ||
          udisks_linux_block_is_unknown_crypto (block);
 }
 
@@ -764,47 +845,13 @@ loop_update (UDisksObject   *object,
 
 /* ---------------------------------------------------------------------------------------------------- */
 
-static void
-free_module_interface_entry (ModuleInterfaceEntry *entry)
-{
-  if (entry->interface != NULL)
-    g_object_unref (entry->interface);
-  g_free (entry);
-}
-
-static void
-ensure_module_ifaces (UDisksLinuxBlockObject *object,
-                      UDisksModuleManager    *module_manager)
-{
-  GList *l;
-  ModuleInterfaceEntry *entry;
-  UDisksModuleInterfaceInfo *ii;
-
-  /* Assume all modules are either unloaded or loaded at the same time, so don't regenerate entries */
-  if (object->module_ifaces == NULL)
-    {
-      object->module_ifaces = g_hash_table_new_full (g_direct_hash, g_direct_equal, NULL, (GDestroyNotify) free_module_interface_entry);
-
-      l = udisks_module_manager_get_block_object_iface_infos (module_manager);
-      for (; l; l = l->next)
-        {
-          ii = l->data;
-          entry = g_new0 (ModuleInterfaceEntry, 1);
-          entry->has_func = ii->has_func;
-          entry->connect_func = ii->connect_func;
-          entry->update_func = ii->update_func;
-          g_hash_table_replace (object->module_ifaces, GSIZE_TO_POINTER (ii->skeleton_type), entry);
-        }
-    }
-}
-
 /**
  * udisks_linux_block_object_uevent:
  * @object: A #UDisksLinuxBlockObject.
  * @action: Uevent action or %NULL
  * @device: A new #UDisksLinuxDevice device object or %NULL if the device hasn't changed.
  *
- * Updates all information on interfaces on @object.
+ * Updates all information on interfaces on @object as a result of incoming uevent processing.
  */
 void
 udisks_linux_block_object_uevent (UDisksLinuxBlockObject *object,
@@ -812,23 +859,25 @@ udisks_linux_block_object_uevent (UDisksLinuxBlockObject *object,
                                   UDisksLinuxDevice      *device)
 {
   UDisksModuleManager *module_manager;
-  GHashTableIter iter;
-  gpointer key;
-  ModuleInterfaceEntry *entry;
+  GList *modules;
+  GList *l;
 
   g_return_if_fail (UDISKS_IS_LINUX_BLOCK_OBJECT (object));
   g_return_if_fail (device == NULL || UDISKS_IS_LINUX_DEVICE (device));
 
   if (device != NULL)
     {
+      g_mutex_lock (&object->device_mutex);
       g_object_unref (object->device);
       object->device = g_object_ref (device);
+      g_mutex_unlock (&object->device_mutex);
       g_object_notify (G_OBJECT (object), "device");
     }
 
   update_iface (UDISKS_OBJECT (object), action, block_device_check, block_device_connect, block_device_update,
                 UDISKS_TYPE_LINUX_BLOCK, &object->iface_block_device);
-  update_iface (UDISKS_OBJECT (object), action, filesystem_check, filesystem_connect, filesystem_update,
+  g_warn_if_fail (object->iface_block_device != NULL);
+  update_iface (UDISKS_OBJECT (object), action, contains_filesystem, filesystem_connect, filesystem_update,
                 UDISKS_TYPE_LINUX_FILESYSTEM, &object->iface_filesystem);
   update_iface (UDISKS_OBJECT (object), action, swapspace_check, swapspace_connect, swapspace_update,
                 UDISKS_TYPE_LINUX_SWAPSPACE, &object->iface_swapspace);
@@ -843,16 +892,47 @@ udisks_linux_block_object_uevent (UDisksLinuxBlockObject *object,
 
   /* Attach interfaces from modules */
   module_manager = udisks_daemon_get_module_manager (object->daemon);
-  if (udisks_module_manager_get_modules_available (module_manager))
+  modules = udisks_module_manager_get_modules (module_manager);
+  for (l = modules; l; l = g_list_next (l))
     {
-      ensure_module_ifaces (object, module_manager);
-      g_hash_table_iter_init (&iter, object->module_ifaces);
-      while (g_hash_table_iter_next (&iter, &key, (gpointer *) &entry))
+      UDisksModule *module = l->data;
+      GType *types;
+
+      types = udisks_module_get_block_object_interface_types (module);
+      for (; types && *types; types++)
         {
-          update_iface (UDISKS_OBJECT (object), action, entry->has_func, entry->connect_func, entry->update_func,
-                        (GType) key, &entry->interface);
+          GDBusInterfaceSkeleton *interface;
+          gboolean keep = TRUE;
+
+          interface = g_hash_table_lookup (object->module_ifaces, GSIZE_TO_POINTER (*types));
+          if (interface != NULL)
+            {
+              /* ask the existing instance to process the uevent */
+              if (udisks_module_object_process_uevent (UDISKS_MODULE_OBJECT (interface),
+                                                       action, object->device, &keep))
+                {
+                  if (! keep)
+                    {
+                      g_dbus_object_skeleton_remove_interface (G_DBUS_OBJECT_SKELETON (object), interface);
+                      g_hash_table_remove (object->module_ifaces, GSIZE_TO_POINTER (*types));
+                    }
+                }
+            }
+          else
+            {
+              /* try create new interface and see if the module is interested in this object */
+              interface = udisks_module_new_block_object_interface (module, object, *types);
+              if (interface)
+                {
+                  /* do coldplug after creation */
+                  udisks_module_object_process_uevent (UDISKS_MODULE_OBJECT (interface), action, object->device, &keep);
+                  g_dbus_object_skeleton_add_interface (G_DBUS_OBJECT_SKELETON (object), interface);
+                  g_warn_if_fail (g_hash_table_replace (object->module_ifaces, GSIZE_TO_POINTER (*types), interface));
+                }
+            }
         }
     }
+  g_list_free_full (modules, g_object_unref);
 }
 
 /* ---------------------------------------------------------------------------------------------------- */
@@ -879,45 +959,58 @@ on_mount_monitor_mount_removed (UDisksMountMonitor  *monitor,
 
 /* ---------------------------------------------------------------------------------------------------- */
 
-
 /**
  * udisks_linux_block_object_trigger_uevent:
  * @object: A #UDisksLinuxBlockObject.
  *
  * Triggers a 'change' uevent in the kernel.
  *
- * The triggered event will bubble up from the kernel through the udev
- * stack and will eventually be received by the udisks daemon process
- * itself. This method does not wait for the event to be received.
+ * Refer to udisks_daemon_util_trigger_uevent() for detailed description.
  */
 void
 udisks_linux_block_object_trigger_uevent (UDisksLinuxBlockObject *object)
 {
-  gchar* path = NULL;
-  gint fd = -1;
+  UDisksLinuxDevice *device;
+  const gchar *sysfs_path;
 
   g_return_if_fail (UDISKS_IS_LINUX_BLOCK_OBJECT (object));
 
-  /* TODO: would be nice with a variant to wait until the request uevent has been received by ourselves */
+  device = udisks_linux_block_object_get_device (object);
+  sysfs_path = g_udev_device_get_sysfs_path (device->udev_device);
+  udisks_daemon_util_trigger_uevent (object->daemon, NULL, sysfs_path);
+  g_object_unref (device);
+}
 
-  path = g_strconcat (g_udev_device_get_sysfs_path (object->device->udev_device), "/uevent", NULL);
-  fd = open (path, O_WRONLY);
-  if (fd < 0)
-    {
-      udisks_warning ("Error opening %s: %m", path);
-      goto out;
-    }
+/**
+ * udisks_linux_block_object_trigger_uevent_sync:
+ * @object: A #UDisksLinuxBlockObject.
+ * @timeout_seconds: Maximum time to wait for the uevent (in seconds).
+ *
+ * Triggers a 'change' uevent in the kernel and waits until it's received and
+ * processed through the uevent queue.
+ *
+ * This is a convenient wrapper around udisks_daemon_util_trigger_uevent_sync().
+ * Refer to this function for detailed documentation.
+ *
+ * Returns: %TRUE if the uevent has been successfully received, %FALSE otherwise
+ * or when the kernel version is too old.
+ */
+gboolean
+udisks_linux_block_object_trigger_uevent_sync (UDisksLinuxBlockObject *object,
+                                               guint                   timeout_seconds)
+{
+  UDisksLinuxDevice *device;
+  const gchar *sysfs_path;
+  gboolean ret;
 
-  if (write (fd, "change", sizeof "change" - 1) != sizeof "change" - 1)
-    {
-      udisks_warning ("Error writing 'change' to file %s: %m", path);
-      goto out;
-    }
+  g_return_val_if_fail (UDISKS_IS_LINUX_BLOCK_OBJECT (object), FALSE);
 
- out:
-  if (fd >= 0)
-    close (fd);
-  g_free (path);
+  device = udisks_linux_block_object_get_device (object);
+  sysfs_path = g_udev_device_get_sysfs_path (device->udev_device);
+  ret = udisks_daemon_util_trigger_uevent_sync (object->daemon, NULL, sysfs_path, timeout_seconds);
+  g_object_unref (device);
+
+  return ret;
 }
 
 /* ---------------------------------------------------------------------------------------------------- */
@@ -936,23 +1029,87 @@ udisks_linux_block_object_trigger_uevent (UDisksLinuxBlockObject *object)
 void
 udisks_linux_block_object_reread_partition_table (UDisksLinuxBlockObject *object)
 {
+  UDisksLinuxDevice *device;
   const gchar *device_file;
   gint fd;
 
   g_return_if_fail (UDISKS_IS_LINUX_BLOCK_OBJECT (object));
 
-  device_file = g_udev_device_get_device_file (object->device->udev_device);
+  device = udisks_linux_block_object_get_device (object);
+  device_file = g_udev_device_get_device_file (device->udev_device);
   fd = open (device_file, O_RDONLY);
   if (fd == -1)
     {
-      udisks_warning ("Error opening %s: %m", device_file);
+      udisks_warning ("Error opening %s while re-reading partition table: %m", device_file);
     }
   else
     {
+      gint num_tries = 0;
+
+      /* acquire an exclusive BSD lock to prevent udev probes.
+       * See also https://systemd.io/BLOCK_DEVICE_LOCKING
+       */
+      while (flock (fd, LOCK_EX | LOCK_NB) != 0)
+        {
+          g_usleep (100 * 1000); /* microseconds */
+          if (num_tries++ > 5)
+            break;
+        }
+
       if (ioctl (fd, BLKRRPART) != 0)
         {
           udisks_warning ("Error issuing BLKRRPART to %s: %m", device_file);
         }
       close (fd);
     }
+
+  g_object_unref (device);
+}
+
+/* ---------------------------------------------------------------------------------------------------- */
+
+/**
+ * udisks_linux_block_object_try_lock_for_cleanup:
+ * @object: A #UDisksLinuxBlockObject.
+ *
+ * Attempts to lock the block device @object for cleanup. Typically used by the cleanup
+ * thread to check whether the block device is busy and cleanup can be performed.
+ *
+ * Returns: %TRUE if lock has been acquired, %FALSE when the block device is busy.
+ */
+gboolean
+udisks_linux_block_object_try_lock_for_cleanup (UDisksLinuxBlockObject *object)
+{
+  g_return_val_if_fail (UDISKS_IS_LINUX_BLOCK_OBJECT (object), FALSE);
+
+  return g_mutex_trylock (&object->cleanup_mutex);
+}
+
+/**
+ * udisks_linux_block_object_lock_for_cleanup:
+ * @object: A #UDisksLinuxBlockObject.
+ *
+ * Acquires the "cleanup" lock of the block device @object. Typically used to signal
+ * the cleanup thread that the block device is busy.
+ */
+void
+udisks_linux_block_object_lock_for_cleanup (UDisksLinuxBlockObject *object)
+{
+  g_return_if_fail (UDISKS_IS_LINUX_BLOCK_OBJECT (object));
+
+  g_mutex_lock (&object->cleanup_mutex);
+}
+
+/**
+ * udisks_linux_block_object_release_cleanup_lock:
+ * @object: A #UDisksLinuxBlockObject.
+ *
+ * Releases a lock previously acquired by udisks_linux_block_object_lock_for_cleanup().
+ */
+void
+udisks_linux_block_object_release_cleanup_lock (UDisksLinuxBlockObject *object)
+{
+  g_return_if_fail (UDISKS_IS_LINUX_BLOCK_OBJECT (object));
+
+  g_mutex_unlock (&object->cleanup_mutex);
 }
