@@ -47,6 +47,7 @@
 #include "udisksmodulemanager.h"
 #include "udiskslinuxfsinfo.h"
 #include "udiskssimplejob.h"
+#include "udisksconfigmanager.h"
 
 /**
  * SECTION:udiskslinuxmanager
@@ -153,6 +154,20 @@ udisks_linux_manager_init (UDisksLinuxManager *manager)
 
   udisks_manager_set_supported_filesystems (UDISKS_MANAGER (manager),
                                             get_supported_filesystems ());
+  udisks_manager_set_supported_encryption_types (UDISKS_MANAGER (manager),
+                                                 get_supported_encryption_types ());
+}
+
+static void
+udisks_linux_manager_constructed (GObject *obj)
+{
+  UDisksLinuxManager *manager = UDISKS_LINUX_MANAGER (obj);
+  UDisksConfigManager *config_manager = udisks_daemon_get_config_manager (manager->daemon);
+
+  udisks_manager_set_default_encryption_type (UDISKS_MANAGER (manager),
+                                              udisks_config_manager_get_encryption (config_manager));
+
+  G_OBJECT_CLASS (udisks_linux_manager_parent_class)->constructed (obj);
 }
 
 static void
@@ -161,6 +176,7 @@ udisks_linux_manager_class_init (UDisksLinuxManagerClass *klass)
   GObjectClass *gobject_class;
 
   gobject_class = G_OBJECT_CLASS (klass);
+  gobject_class->constructed  = udisks_linux_manager_constructed;
   gobject_class->finalize     = udisks_linux_manager_finalize;
   gobject_class->set_property = udisks_linux_manager_set_property;
   gobject_class->get_property = udisks_linux_manager_get_property;
@@ -369,7 +385,7 @@ handle_loop_setup (UDisksManager          *object,
       g_dbus_method_invocation_return_error (invocation,
                                              UDISKS_ERROR,
                                              UDISKS_ERROR_FAILED,
-                                             "Error determing path: %m");
+                                             "Error determining path: %m");
       goto out;
     }
   path[path_len] = '\0';
@@ -414,16 +430,20 @@ handle_loop_setup (UDisksManager          *object,
   error = NULL;
   wait_data.loop_device = loop_device;
   wait_data.path = path;
+  udisks_daemon_util_trigger_uevent_sync (manager->daemon,
+                                          loop_device,
+                                          NULL,
+                                          UDISKS_DEFAULT_WAIT_TIMEOUT);
   loop_object = udisks_daemon_wait_for_object_sync (manager->daemon,
                                                     wait_for_loop_object,
                                                     &wait_data,
                                                     NULL,
-                                                    10, /* timeout_seconds */
+                                                    UDISKS_DEFAULT_WAIT_TIMEOUT,
                                                     &error);
   if (loop_object == NULL)
     {
       g_prefix_error (&error,
-                      "Error waiting for loop object after creating %s",
+                      "Error waiting for loop object after creating '%s': ",
                       loop_device);
       g_dbus_method_invocation_take_error (invocation, error);
       goto out;
@@ -647,6 +667,7 @@ handle_mdraid_create (UDisksManager         *_object,
                                                  UDISKS_ERROR_FAILED,
                                                  "Object path %s for index %u is not a block device",
                                                  arg_blocks[n], n);
+          g_object_unref (object);
           success = FALSE;
           goto out;
         }
@@ -661,6 +682,8 @@ handle_mdraid_create (UDisksManager         *_object,
                                                  "Error opening device %s while creating mdraid: %m",
                                                  device_file);
           g_free (device_file);
+          g_object_unref (block);
+          g_object_unref (object);
           success = FALSE;
           goto out;
         }
@@ -684,7 +707,7 @@ handle_mdraid_create (UDisksManager         *_object,
           else
             {
               g_prefix_error (&error,
-                              "Error wiping device %s to be used in the RAID array:",
+                              "Error wiping device '%s' to be used in the RAID array: ",
                               udisks_block_get_device (block));
               g_dbus_method_invocation_take_error (invocation, error);
               success = FALSE;
@@ -720,7 +743,7 @@ handle_mdraid_create (UDisksManager         *_object,
 
   if (!bd_md_create (array_name, arg_level, disks, 0, NULL, FALSE, arg_chunk, NULL, &error))
     {
-      g_prefix_error (&error, "Error creating RAID array:");
+      g_prefix_error (&error, "Error creating RAID array: ");
       g_dbus_method_invocation_take_error (invocation, error);
       udisks_simple_job_complete (UDISKS_SIMPLE_JOB (job), FALSE, error->message);
       success = FALSE;
@@ -733,28 +756,31 @@ handle_mdraid_create (UDisksManager         *_object,
       raid_node = bd_md_node_from_name (array_name, &error);
       if (!raid_node)
         {
-          g_prefix_error (&error, "Failed to get md node for array %s", array_name);
+          g_prefix_error (&error, "Failed to get md node for array '%s': ", array_name);
           g_dbus_method_invocation_take_error (invocation, error);
           success = FALSE;
           goto out;
         }
       raid_device_file = g_strdup_printf ("/dev/%s", raid_node);
     }
-
   else
     raid_device_file = g_strdup (array_name);
 
+  /* trigger uevent on the newly created md node */
+  udisks_daemon_util_trigger_uevent_sync (manager->daemon, raid_device_file, NULL,
+                                          UDISKS_DEFAULT_WAIT_TIMEOUT);
+
   /* ... then, sit and wait for raid array object to show up */
   array_object = udisks_daemon_wait_for_object_sync (manager->daemon,
                                                      wait_for_array_object,
                                                      raid_device_file,
                                                      NULL,
-                                                     10, /* timeout_seconds */
+                                                     UDISKS_DEFAULT_WAIT_TIMEOUT,
                                                      &error);
   if (array_object == NULL)
     {
       g_prefix_error (&error,
-                      "Error waiting for array object after creating %s",
+                      "Error waiting for array object after creating '%s': ",
                       raid_device_file);
       g_dbus_method_invocation_take_error (invocation, error);
       success = FALSE;
@@ -795,7 +821,7 @@ handle_mdraid_create (UDisksManager         *_object,
         g_clear_error (&error);
       else
         {
-          g_prefix_error (&error, "Error wiping raid device %s:", raid_device_file);
+          g_prefix_error (&error, "Error wiping raid device '%s': ", raid_device_file);
           g_dbus_method_invocation_take_error (invocation, error);
           success = FALSE;
           goto out;
@@ -849,37 +875,117 @@ handle_mdraid_create (UDisksManager         *_object,
 
 /* ---------------------------------------------------------------------------------------------------- */
 
-static void
-load_modules (UDisksDaemon *daemon)
+typedef struct
 {
+  UDisksManager         *object;
+  GDBusMethodInvocation *invocation;
+  gchar                 *module_name;
+} EnableModulesData;
+
+static gboolean
+load_modules_in_idle_cb (gpointer user_data)
+{
+  EnableModulesData *data = user_data;
+  UDisksLinuxManager *manager = UDISKS_LINUX_MANAGER (data->object);
   UDisksModuleManager *module_manager;
 
-  g_return_if_fail (UDISKS_IS_DAEMON (daemon));
+  module_manager = udisks_daemon_get_module_manager (manager->daemon);
+  if (data->module_name == NULL)
+    {
+      /* Load all modules */
+      udisks_module_manager_load_modules (module_manager);
+      /* Avoid deprecation warning from udisks_manager_complete_enable_modules() */
+      g_dbus_method_invocation_return_value (data->invocation, g_variant_new ("()"));
+    }
+  else
+    {
+      GError *error = NULL;
 
-  module_manager = udisks_daemon_get_module_manager (daemon);
-  udisks_module_manager_load_modules (module_manager);
+      /* Load single requested module */
+      if (! udisks_module_manager_load_single_module (module_manager, data->module_name, &error))
+        {
+          g_prefix_error (&error, "Error initializing module '%s': ", data->module_name);
+          g_warning ("%s", error->message);
+          g_dbus_method_invocation_take_error (data->invocation, error);
+        }
+      else
+        {
+          udisks_manager_complete_enable_module (data->object, data->invocation);
+        }
+    }
+  g_object_unref (data->object);
+  g_object_unref (data->invocation);
+  g_free (data->module_name);
+  g_free (data);
+
+  return G_SOURCE_REMOVE;
 }
 
 static gboolean
-handle_enable_modules (UDisksManager *object,
+handle_enable_modules (UDisksManager         *object,
                        GDBusMethodInvocation *invocation,
-                       gboolean arg_enable)
+                       gboolean               arg_enable)
 {
   UDisksLinuxManager *manager = UDISKS_LINUX_MANAGER (object);
+  EnableModulesData *data;
 
   if (! arg_enable)
     {
       /* TODO: implement proper module unloading */
       g_dbus_method_invocation_return_error_literal (invocation,
                                                      G_IO_ERROR, G_IO_ERROR_INVALID_ARGUMENT,
-                                                     "Invalid value \"FALSE\"");
+                                                     "Module unloading is not currently supported.");
       return TRUE;
     }
 
   if (! udisks_daemon_get_disable_modules (manager->daemon))
-    load_modules (manager->daemon);
+    {
+      data = g_new0 (EnableModulesData, 1);
+      data->object = g_object_ref (object);
+      data->invocation = g_object_ref (invocation);
+      /* push to idle, process in main thread */
+      g_idle_add (load_modules_in_idle_cb, data);
+    }
 
-  udisks_manager_complete_enable_modules (object, invocation);
+  return TRUE; /* returning TRUE means that we handled the method invocation */
+}
+
+static gboolean
+handle_enable_module (UDisksManager         *object,
+                      GDBusMethodInvocation *invocation,
+                      const gchar           *arg_name,
+                      gboolean               arg_enable)
+{
+  UDisksLinuxManager *manager = UDISKS_LINUX_MANAGER (object);
+  EnableModulesData *data;
+
+  if (! udisks_module_validate_name (arg_name))
+    {
+      g_dbus_method_invocation_return_error (invocation,
+                                             G_IO_ERROR, G_IO_ERROR_INVALID_ARGUMENT,
+                                             "Requested module name '%s' is not a valid udisks2 module name.",
+                                             arg_name);
+      return TRUE;
+    }
+
+  if (! arg_enable)
+    {
+      /* TODO: implement proper module unloading */
+      g_dbus_method_invocation_return_error_literal (invocation,
+                                                     G_IO_ERROR, G_IO_ERROR_INVALID_ARGUMENT,
+                                                     "Module unloading is not currently supported.");
+      return TRUE;
+    }
+
+  if (! udisks_daemon_get_disable_modules (manager->daemon))
+    {
+      data = g_new0 (EnableModulesData, 1);
+      data->object = g_object_ref (object);
+      data->invocation = g_object_ref (invocation);
+      data->module_name = g_strdup (arg_name);
+      /* push to idle, process in main thread */
+      g_idle_add (load_modules_in_idle_cb, data);
+    }
 
   return TRUE; /* returning TRUE means that we handled the method invocation */
 }
@@ -1065,6 +1171,7 @@ handle_get_block_devices (UDisksManager         *object,
                                               invocation,
                                               block_paths);
 
+  g_free (block_paths);
   g_slist_free_full (blocks, g_object_unref);
 
   return TRUE;  /* returning TRUE means that we handled the method invocation */
@@ -1073,7 +1180,7 @@ handle_get_block_devices (UDisksManager         *object,
 static gboolean
 compare_paths (UDisksManager         *object,
                UDisksBlock           *block,
-               gchar                 *path)
+               const gchar           *path)
 {
   const gchar *const *symlinks = NULL;
 
@@ -1097,10 +1204,9 @@ handle_resolve_device (UDisksManager         *object,
                        GVariant              *arg_devspec,
                        GVariant              *arg_options)
 {
-
-  gchar *devpath = NULL;
-  gchar *devuuid = NULL;
-  gchar *devlabel = NULL;
+  const gchar *devpath = NULL;
+  const gchar *devuuid = NULL;
+  const gchar *devlabel = NULL;
 
   GSList *blocks = NULL;
   GSList *blocks_p = NULL;
@@ -1114,9 +1220,9 @@ handle_resolve_device (UDisksManager         *object,
   gboolean found = FALSE;
   guint i = 0;
 
-  g_variant_lookup (arg_devspec, "path", "s", &devpath);
-  g_variant_lookup (arg_devspec, "uuid", "s", &devuuid);
-  g_variant_lookup (arg_devspec, "label", "s", &devlabel);
+  g_variant_lookup (arg_devspec, "path", "&s", &devpath);
+  g_variant_lookup (arg_devspec, "uuid", "&s", &devuuid);
+  g_variant_lookup (arg_devspec, "label", "&s", &devlabel);
 
   blocks = get_block_objects (object, &num_blocks);
 
@@ -1132,26 +1238,26 @@ handle_resolve_device (UDisksManager         *object,
 
       if (found)
         {
-          ret = g_slist_prepend (ret, g_object_ref (blocks_p->data));
+          ret = g_slist_prepend (ret, blocks_p->data);
           num_found++;
         }
     }
 
-    ret_paths = g_new0 (const gchar *, num_found + 1);
-    for (i = 0,ret_p = ret; ret_p != NULL; ret_p = ret_p->next, i++)
-      {
-        ret_paths[i] = g_dbus_object_get_object_path (g_dbus_interface_get_object (G_DBUS_INTERFACE (ret_p->data)));
-      }
+  ret_paths = g_new0 (const gchar *, num_found + 1);
+  for (i = 0,ret_p = ret; ret_p != NULL; ret_p = ret_p->next, i++)
+    {
+      ret_paths[i] = g_dbus_object_get_object_path (g_dbus_interface_get_object (G_DBUS_INTERFACE (ret_p->data)));
+    }
 
-    udisks_manager_complete_resolve_device (object,
-                                            invocation,
-                                            ret_paths);
+  udisks_manager_complete_resolve_device (object,
+                                          invocation,
+                                          ret_paths);
 
-    g_slist_free_full (blocks, g_object_unref);
-    g_slist_free_full (ret, g_object_unref);
-
-    return TRUE;  /* returning TRUE means that we handled the method invocation */
+  g_free (ret_paths);
+  g_slist_free_full (blocks, g_object_unref);
+  g_slist_free (ret);
 
+  return TRUE;  /* returning TRUE means that we handled the method invocation */
 }
 
 /* ---------------------------------------------------------------------------------------------------- */
@@ -1162,6 +1268,7 @@ manager_iface_init (UDisksManagerIface *iface)
   iface->handle_loop_setup = handle_loop_setup;
   iface->handle_mdraid_create = handle_mdraid_create;
   iface->handle_enable_modules = handle_enable_modules;
+  iface->handle_enable_module = handle_enable_module;
   iface->handle_can_format = handle_can_format;
   iface->handle_can_resize = handle_can_resize;
   iface->handle_can_check = handle_can_check;
